/* CA utility provider: automatic certificate request signer.

   For each open request:

       If the requesting user is registered in the netacts database
       as the owner of the certificate's domain and:

         1)  the requestor is approved for auto-sign, or

         2)  the request is for a small web server, or

         3)  the request is for a client or server certificate, or

         4)  the request is a renewal

       We will sign the request.

       Small means: (UW users < 193) && (non-UW users < 37)

   */


#include <stdio.h>
#include <signal.h>
#include <string.h>

#include "webtpl.h"

#include "db.h"
#include "util.h"
#include "cassl.h"
#include "ca.h"


extern int errno;

#define MAXUW 139
#define MAXEX  37

char *errstr = NULL;
static char *remote_user = "somebody";

#define U_SIGNED_MSG \
 "Your certificate request for %s (No. %d) has been signed. \n\n" \
 "Go to https://certs.cac.washington.edu/services/?req=full&rno=%d to retrieve it.\n\n"


/* Compute lifetime of the new cert.  If this is a renewal
   count how many days for 1-year from old cert's expiration */

#define DEFAULT_DAYS 366
#define DAY   (24*60*60)

int count_the_days(SvrReq R)
{
   SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
   SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));

   time_t now = time(NULL);
   time_t basetime = now;
   struct tm t;

   if (R->crtno) {

      /* reference an existing good cert */
      db_select_crt(R->crtno, 0, CRT_OK);
      if (db_get_crt(S)) {
          basetime = S->expire_date;
          printf("using existing cert, base=%d\n", basetime);
      } 
      db_free_crt(); 

   } else {

      db_select_notes(R->reqno);
      while (db_get_notes(N)) {
         int rn, n;
         n = sscanf(N->notes, "Renewal of request %d by", &rn);
         if (n==1 && rn>1) {
            db_select_crt(0, rn, CRT_OK);
            if (db_get_crt(S)) {
               basetime = S->expire_date;
               printf("using reference from %d, base=%d\n", rn, basetime);
            }
            db_free_crt();
         }
      }
      db_free_notes(); 
   }
   free(S);
   free(N);

   /* find 3-year from base */
   localtime_r(&basetime, &t);
   t.tm_year += 3;
   // t.tm_year += 1;
   printf("expire is: %d/%d/%d %d:%d:%d \n", t.tm_mon+1, t.tm_mday, t.tm_year+1900, t.tm_hour, t.tm_min, t.tm_sec);
   return (((mktime(&t)-now)/DAY)+1);

}

/* sign a request */

int sign_req(SvrReq R, int key_usage)
{
   SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));

   int rc;
   char *ret = NULL;
   char *cert = NULL;
   char *dn;
   int serial;
   time_t issuedt;
   int crtno = 0;
   X509 *xcrt;
   int flags = CA_FL_ALTNAMDNS;
   int ndays = 0;
   int ext = 0;   // default is no keyusage flags
   
   ndays = count_the_days(R);
   if (key_usage&0xf0) ext = CA_EXT_SC;  // includes signature and encipherment usage keys
   
   rc = ca_cmd(CA_SIGN, remote_user, R->request, ndays*100 + ext, flags, &ret);

   if ((rc==0) && ret) {
      cert = strstr(ret,"-----BEGIN CERTIFICATE-----");
      if (cert) {

         xcrt = ssl_txt2crt(ret);
         S->reqno = R->reqno;
         strcpy(S->cert, cert);
         strcpy(S->issue_by, remote_user);
         S->casn = ssl_get_sn(xcrt);
         S->issue_date = ssl_get_issue(xcrt);
         S->expire_date = ssl_get_expire(xcrt);
             
         crtno = db_load_crt(S);
         if (crtno) {
            char *ssub = (char*) malloc(strlen(R->dn)+48);
            char *umsg = (char*) malloc(strlen(U_SIGNED_MSG) + strlen(R->dn) + 64);

            db_set_req_crtno(R->reqno, crtno);

            /* send the user a notice */
            sprintf(ssub,"Certificate request %d (%s) signed", R->reqno, R->dn); 
            sprintf(umsg,U_SIGNED_MSG, R->dn, R->reqno, R->reqno);
            send_owner_mail(R, ssub, umsg);
            free(umsg);

            /* send the list a notice */
            sprintf(ssub,"request %d signed by %s\n", R->reqno, remote_user); 
            if (strcmp(R->reqid,"fox")) send_mail(GETS_MAIL, ssub,
                  "That's all.\n");

            free(ssub);
            write_log(SERVICES_LOG, remote_user, "sign", R->reqno, crtno);

         } else errstr = ret;
      }
   } else errstr = ret;

   return (crtno);
}



main(int argc, char **argv)
{
  char *user;
  char *seqstr;
  int seq = 0;
  int s;
  char *p;
  char *msgstr = NULL;
  int reqno;
  int us;

  SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
  SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));
  X509_REQ *xreq;
  int rc;
  char smsg[24];
  char *dnmsg;
  char *altmsg;
  time_t now = time(NULL);
  char tmsg[64];
  int nuw, nex;

  strcpy(tmsg, ctime(&now));
  if (p=strchr(tmsg,'\n')) *p = '\0';

  db_connect();

  db_select_req(0, NULL, NULL, RST_ACTIVE, NULL);

  while (db_get_req(R)>0) {
    int ok2sign = 0;
    int usage = 0;

    if (R->status==ST_REVREQ) continue; /* we don't touch these */
    xreq = ssl_txt2req(R->request);
    dnmsg = ssl_chk_req_dn(xreq, R->reqid);
    altmsg = ssl_chk_req_altnames(xreq, R->reqid);
    usage = ssl_get_keyusage_req(xreq);

    us = db_get_user_status(R->reqid);

    if ((!dnmsg) && (!altmsg)) {
       if ((us==US_APPROVED) || (R->status==ST_RENEW)) ok2sign = 1; /* sign */
       else {
           /* Maybe is small request - see top of program */
           /* Note the exact match with notes form text. */
           db_select_notes(R->reqno);
           while ((ok2sign==0) && db_get_notes(N)) {
              if (!strncmp(N->notes, "SSL Client", 10)) ok2sign=2; 
              if (!strncmp(N->notes, " SSL Server", 11)) ok2sign=3; 
              if (!strncmp(N->notes, "Web Server", 10)) {
                 p = strchr(N->notes, ':');
                 if (p) nuw = atoi(p+1);
                 p = strchr(N->notes, ',');
                 if (p) nex = atoi(p+1);
                 if ((nuw+nex) && (nuw<=MAXUW) && (nex<=MAXEX)) ok2sign = 4;
              }
           }
           db_free_notes();
       }
    }

    if (ok2sign) { 
       printf("%s signing(%d) %d [%d], usage=%x, from %s\n", tmsg,ok2sign,R->dn, R->reqno, usage, R->reqid);
       rc = sign_req(R, usage); 
       if (!rc) printf("%s: CA says \"%s\"\n", tmsg, errstr);
    } else {           
       printf("%s skipping %d, from %s (%s)\n", tmsg, R->reqno, R->reqid,
          dnmsg?dnmsg:altmsg?altmsg:"");
    }
  }
}

