/* CA database */

#ifndef _ca_db_h_
#define _ca_db_h_

#include <time.h>

/* These status are in the DB request record */

/* status */
#define ST_NONE     0
#define ST_REQ      1    /* cert requested 1st time - no cert yet */
#define ST_ISSUE    2    /* db contains valid, active cert */
#define ST_DENY     3    /* request was denied - no cert */
#define ST_DELETED  4    /* a denied request was deleted - no cert */
#define ST_REVOKED  5    /* a cert was revoked - db contains revoked cert */
#define ST_RENEW    6    /* renewal requested for good cert */
#define ST_EXPIRED  7    /* last cert has expired */
#define ST_CANCEL   8    /* a request was cancelled by user - no cert */
#define ST_REVREQ   9    /* revocation requested */

/* request status combinations */
#define RST_ACTIVE   101 /* needs admin attention: REQ|RENEW|REVREQ */
#define RST_ALIVE    102 /* can be 'gotten': ISSUE|RENEW */
#define RST_CANBLE   103 /* can be cancelled: REQ|RENEW */


/* These status are in the cert record. */

#define CRT_OK       0    /* ok */
#define CRT_REVOKED  1    /* revoked */
#define CRT_EXPIRED  2    /* expired */
#define CRT_EXPIRING 3    /* expiring (<30 days) */

#ifdef UWCA_db_lib
char *req_status_text[] = {
   "none",
   "pending",
   "issued",
   "denied",
   "deleted",
   "revoked",
   "renew",
   "expired",
   "cancelled",
   "revoke request"
};
char *crt_status_text[] = {
   "issued",
   "revoked",
   "expired",
   "expiring"
};
#endif


/* sort keys */

#define SR_RNO 1     /* reqno */
#define SR_RDN 2     /* DN */
#define SR_RID 3     /* user */
#define SR_RRD 4     /* req date */
#define SR_RST 5     /* status */
#define SR_RIS 6     /* issue date (from svcrt) */
#define SR_REX 7     /* expire date (from svcrt) */
#define SR_NUM 7

#ifdef UWCA_db_lib
char *sort_vars[] = {
  "void",
  "svreq.reqno",
  "svreq.dn",
  "svreq.reqid",
  "svreq.request_date",
  "svreq.status",
  "svcrt.issue_date",
  "svcrt.expire_date"
};
#endif

/* requests table */
typedef struct SvrReq__ {
  int  reqno;
  char reqid[24];        /* requestor netid */
  int  status;
  time_t  request_date;
  int  crtno;            /* last issued cert */
  char dn[257];          /* DN from the request */
  char request[16384];
  int hash;
} SvrReq_, *SvrReq;


/* certs table */
typedef struct SvrCert__ {
  int  crtno;
  int  reqno;           /* link to the request */
  int  casn;            /* sn from the CA */
  int  status;          /* in db or computed */
  time_t issue_date;
  char issue_by[16];
  char cert[16384];
  time_t revoke_date;
  char revoke_by[16];
  time_t expire_date;
  int expno;
} SvrCert_, *SvrCert;


/* notes table */
typedef struct SvrNotes__ {
  int notno;
  int reqno;
  char notes[16384];
} SvrNotes_, *SvrNotes;


/* users table */

#define US_UNKNOWN   0
#define US_KNOWN     1
#define US_APPROVED  2
#define US_BANNED    3

typedef struct SvrUser__ {
  char id[24];
  int status;
} SvrUser_, *SvrUser;

/* groups table */
typedef struct SvrGrp__ {
  char gid[24];
  char auth_group[1024];
  char email[1024];
} SvrGrp_, *SvrGrp;

/* usr-grp table */
typedef struct SvrUG__ {
  char id[12];
  char gid[24];
} SvrUG_, *SvrUG;

/* owners table */
typedef struct SvrOwn__ {
  int reqno;
  char id[12];
} SvrOwn_, *SvrOwn;


/* prototypes */

int db_connect();
void db_disconnect();
char *db_last_error();
int db_get_req(SvrReq R);
int db_get_crt(SvrCert S);
int db_get_notes(SvrNotes N);
int db_get_user(SvrUser U);
int db_select_req(int reqno, char *dn, char *id, int status, char *srtkey);
int db_select_crt(int crtno, int reqno, int status);
int db_select_notes(int reqno);
int db_select_user(char *id);
int db_load_req(SvrReq R);
int db_load_notes(int reqno, char *notes);
int db_load_user(char *id, int status);
int db_set_req_status(int reqno, int status, char *id);
int db_set_crt_status(int crtno, int status, int reqno);
int db_set_crt_revoke_date(int crtno, time_t rdate);
int db_set_req_reqid(int reqno, char *oldid, char *newid);
int db_set_user_status(char *id, int status);
int db_delete_req(int reqno);
int db_get_user_status(char *id);
extern char *req_status_text[];
extern char *crt_status_text[];

int db_update_owners(SvrReq R);


#endif
