/* Generate a CRL */

#include <sys/file.h>
#include <fcntl.h>

#include "caserver.h"

/* show the details of an error. These go to the local log. */
static void show_error (const char *file, int lineno, const char *msg)
{
  fprintf (stderr, "** %s:%d %s\n", file, lineno, msg);
  ERR_print_errors_fp (stderr);
}

/* These are usually bad requests */
#define ERR(msg) {show_error(__FILE__, __LINE__, msg);rettxt=msg;goto err;}
/* These are internal and should never happen */
#define INTERR() ERR("INT");



/* ---- Extensions ---- */

typedef struct ExtTxt__ {
  char *key;
  char *value;
} ExtTxt_, *ExtTxt;


/* Default CRL extensions */
static ExtTxt_ def_ext[] = {
  {"authorityKeyIdentifier", "keyid:always,issuer:always"},
  {"issuerAltName", "issuer:copy"},
  {NULL,NULL}
};


static void seed_prng(void)
{
  RAND_load_file(CA_ROOT "etc/rand", 1024);
}

/* Generate the CRL */

int ssl_gencrl(char *keypass, char **errtxt)
{
  int i, subjAltName_pos;
  int isn;
  long serial = 0;
  EVP_PKEY *pkey, *CApkey;
  X509_CRL *crl;
  X509_CRL_INFO *ci;
  const EVP_MD *digest;
  X509 *cert, *CAcert;
  X509_REQ *req;
  X509_NAME *name;
  X509V3_CTX ctx;
  X509_EXTENSION *subjAltName;
  STACK_OF (X509_EXTENSION) * req_exts;
  BIO *in, *out;
  FILE *fp;
  ExtTxt exttxt = def_ext;
  BUF_MEM *outbuf = NULL;
  int ret = 0;
  int crl_v = 02;
  long crldays = 30;
  X509_REVOKED *rev;
  ASN1_TIME *atm;
  ASN1_INTEGER *asn;
  char *rettxt;
  int seq;
  time_t now = time(NULL);
  struct tm *t = localtime(&now);

  int crtdb;
  CertRec C;

  OpenSSL_add_all_algorithms ();
  ERR_load_crypto_strings ();
  seed_prng ();
  

  /* ---  get the CA cert and key ------ */

  if (!(fp = fopen (CA_CERT, "r"))) INTERR ();
  if (!(CAcert = PEM_read_X509 (fp, NULL, NULL, NULL))) INTERR();
  fclose (fp);

  if (!(fp = fopen (CA_KEY, "r"))) INTERR();
  if (!(CApkey = PEM_read_PrivateKey (fp, NULL, NULL, keypass))) INTERR();
  fclose (fp);



  /* --- Build the CRL --------- */


  if (!(crl = X509_CRL_new())) ERR("newcrl");
  ci = crl->crl;
  if (!X509_CRL_set_issuer_name(crl, X509_get_subject_name(CAcert))) ERR("issuer");
  if (!X509_CRL_set_version(crl, 1)) ERR("version");

  /* times */

  atm = ASN1_TIME_new();
  if (!atm) ERR("time");
  X509_gmtime_adj(atm,0);
  X509_CRL_set_lastUpdate(crl, atm);
  X509_gmtime_adj(atm,(crldays*24)*60*60);
  X509_CRL_set_nextUpdate(crl, atm);
  ASN1_TIME_free(atm);

  /* add revoked certs */

  crtdb = open(CA_CERTDB, O_CREAT|O_RDWR, 0660);
  if (!crtdb) ERR("crtdb");

  C = (CertRec) malloc(sizeof(CertRec_));

  isn = 0;
  seq = 1;
  while (read(crtdb, C, sizeof(CertRec_))==sizeof(CertRec_)) {
     time_t rt = time(NULL);

     if ((isn++!=C->casn) && (C->casn!=0)) ERR("sn mismatch");
     if (!C->revoke_date) continue;
     if (C->expire_date<now) continue;   /* old cert */
    
     if (!(rev=X509_REVOKED_new())) ERR("newrev");
     atm = ASN1_TIME_set(NULL, C->revoke_date);
     if (!X509_REVOKED_set_revocationDate(rev, atm)) ERR("revdate");
     asn = ASN1_INTEGER_new();
     ASN1_INTEGER_set(asn, C->casn);
     X509_REVOKED_set_serialNumber(rev, asn);
/**
     rev->extensions = NULL;
     rev->sequence = seq++;
     sk_X509_REVOKED_push(ci->revoked,rev);
  **/
     X509_CRL_add0_revoked(crl,rev);
     seq++;
  }
  close(crtdb);
  X509_CRL_sort(crl);

#ifdef INCLUDE_EXTENSIONS
  /* set extensions */
  X509V3_set_ctx (&ctx, CAcert, NULL, NULL, NULL, 0);
  while (exttxt->key) {
      X509_EXTENSION *ext;
      if (!(ext=X509V3_EXT_conf(NULL,&ctx,exttxt->key, exttxt->value))) {
	  fprintf (stderr, "INT: making ext \"%s = %s\"\n",
		   exttxt->key, exttxt->value);
	  INTERR();
	}
      if (!X509_add_ext (cert, ext, -1))
	{
	  fprintf (stderr, "INT: adding ext \"%s = %s\"\n",
		   exttxt->key, exttxt->value);
	  INTERR();
	}
      X509_EXTENSION_free (ext);
      exttxt++;
  }
#endif

  /* sign */

  if (EVP_PKEY_type (CApkey->type) == EVP_PKEY_DSA) digest = EVP_dss1 ();
  else if (EVP_PKEY_type (CApkey->type) == EVP_PKEY_RSA) digest = EVP_sha1 ();
  else digest = EVP_md5();
  /* else ERR("digest"); */
  if (!X509_CRL_sign(crl,CApkey,digest)) ERR("sign");

  /* output the crl to a local file */

   out = BIO_new(BIO_s_file());
   if (out == NULL) ERR("out");

   if (BIO_write_filename(out,CRL_NAME)<=0) ERR("writefn");;

   /* write a log entry */

   fp = fopen(CA_LOG,"a");
   if (fp) {
      fprintf(fp,"%4d/%02d/%02d %02d:%02d:%02d newcrl\n",
            t->tm_year+1900, t->tm_mon+1, t->tm_mday,
            t->tm_hour, t->tm_min, t->tm_sec);
      fclose(fp);
   }

   ret=PEM_write_bio_X509_CRL(out, crl);
   BIO_free_all(out);

   if (errtxt) *errtxt = NULL;
   return (1);

err:
   if (errtxt) *errtxt = rettxt;
   return (0);
}
