/* Sign a request */

#include <sys/file.h>
#include <fcntl.h>

#include "caserver.h"
#include <openssl/pkcs12.h>

/* show the details of an error. These go to the local log. */
void show_error (const char *file, int lineno, const char *msg)
{
  fprintf (stderr, "** %s:%d %s\n", file, lineno, msg);
  ERR_print_errors_fp (stderr);
}

/* These are usually bad requests */
#define ERR(msg) {show_error(__FILE__, __LINE__, msg);rettxt=strdup(msg);goto err;}
/* These are internal and should never happen */
#define INTERR() ERR("INT");

/* ---- Extensions ---- */

typedef struct ExtTxt__ {
  char *key;
  char *value;
} ExtTxt_, *ExtTxt;


/* Default extensions (client-server) */
ExtTxt_ def_ext[] = {
  {"basicConstraints", "critical,CA:FALSE"},
  {"extendedKeyUsage", "clientAuth,serverAuth"},
  {"subjectKeyIdentifier", "hash"},
  {"subjectAltName", "-cn-"},
  {"authorityKeyIdentifier", "keyid,issuer:always"},
  {"crlDistributionPoints",
         "URI:http://certs.cac.washington.edu/UWServicesCA.crl"},
  {NULL,NULL}
};

/* client-server cert extensions */
ExtTxt_ cs_ext[] = {
  {"basicConstraints", "critical,CA:FALSE"},
  {"keyUsage", "digitalSignature,nonRepudiation,keyEncipherment,dataEncipherment"},
  {"extendedKeyUsage", "clientAuth,serverAuth"},
  {"subjectKeyIdentifier", "hash"},
  {"subjectAltName", "-cn-"},
  {"authorityKeyIdentifier", "keyid,issuer:always"},
  {"crlDistributionPoints",
         "URI:http://certs.cac.washington.edu/UWServicesCA.crl"},
  {NULL,NULL}
};

/* code-signer extensions */
ExtTxt_ code_ext[] = {
  {"basicConstraints", "critical,CA:FALSE"},
  {"extendedKeyUsage", "codeSigning"},
  {"subjectKeyIdentifier", "hash"},
  {"subjectAltName", "-cn-"},
  {"authorityKeyIdentifier", "keyid,issuer:always"},
  {"crlDistributionPoints",
         "URI:http://certs.cac.washington.edu/UWServicesCA.crl"},
  {NULL,NULL}
};


/* ---- Cert db ----- */

static int update_db(X509 *crt)
{
  int crtdb = open(CA_CERTDB, O_CREAT|O_WRONLY, 0660);
  CertRec C = (CertRec) malloc(sizeof(CertRec_));
  int sn;
  int nb;
  char *dn;
  FILE *log;
  time_t now = time(NULL);
  struct tm *t = localtime(&now);

  if (!crtdb) return (0);
 
  memset(C, '\0', sizeof(CertRec_));

  sn = i2i_ASN1_INTEGER(crt->cert_info->serialNumber);
  C->casn = sn;
  C->issue_date = ASN1_UTCTIME_get(X509_get_notBefore(crt));
  C->expire_date = ASN1_UTCTIME_get(X509_get_notAfter(crt));
  C->revoke_date = 0;
  dn = X509_NAME_oneline(X509_get_subject_name(crt),NULL,0);
  dn[239] = '\0';
  strcpy(C->dn, dn);
  
  if (lseek(crtdb, sn*sizeof(CertRec_), SEEK_SET)>0) {
     nb = write(crtdb, C, sizeof(CertRec_));
  }
  close(crtdb);
  free (C);

  /* write log */

  log = fopen(CA_LOG,"a");
  if (log) {
     fprintf(log,"%4d/%02d/%02d %02d:%02d:%02d sn=%6d(%x): %s \n",
           t->tm_year+1900, t->tm_mon+1, t->tm_mday,
           t->tm_hour, t->tm_min, t->tm_sec,
           sn, sn, dn);
     fclose(log);
  }

  free(dn);

  return (nb);
}



/* Sign a request (reqtxt), returns sn, (cert in crttxt) 
   reqext = cert length (days) * 100 + cert type code */

int ssl_sign(char *reqtxt, int reqext, int flags, char *keypass, char **crttxt)
{
  char *rettxt = NULL;
  int i, subjAltName_pos;
  long serial = 0;
  EVP_PKEY *pkey, *CApkey;
  const EVP_MD *digest;
  X509 *cert, *CAcert;
  X509_REQ *req;
  X509_NAME *name;
  X509_NAME_ENTRY *nent;
  ASN1_STRING *nstr;
  X509V3_CTX ctx;
  STACK_OF (X509_EXTENSION) * req_exts;
  ASN1_OBJECT *obj;
  BIO *in, *out;
  FILE *fp;
  ExtTxt exttxt;
  BUF_MEM *outbuf = NULL;
  int ret = 0;
  char *name_cn;
  char *alt_names;
  X509_EXTENSION *ext;
  int reqlt = reqext/100; /* lifetime in days */
  int reqtype = reqext - (reqlt*100);

  OpenSSL_add_all_algorithms ();
  ERR_load_crypto_strings ();
  seed_prng ();

  
  /* --- Request --------- */

   /* Use default if not specifically asking for something special */

   if (reqtype==CA_EXT_SC) exttxt = cs_ext;
   else if (reqtype==CA_EXT_COD) exttxt = code_ext;
   else exttxt = def_ext;

   in = BIO_new_mem_buf(reqtxt, strlen(reqtxt));
   if (in == NULL) INTERR();

   BIO_set_close(in, BIO_NOCLOSE);
   req=PEM_read_bio_X509_REQ(in,NULL,NULL,NULL);
   /* CLOSE */


  /* verify the request */

  if (!(pkey = X509_REQ_get_pubkey (req))) ERR("REQ: Bad request");

  if (X509_REQ_verify (req, pkey) != 1) ERR("REQ: Bad request sig");


  /* ---  get the CA cert and key ------ */

  if (!(fp = fopen (CA_CERT, "r"))) INTERR ();
  if (!(CAcert = PEM_read_X509 (fp, NULL, NULL, NULL))) INTERR();
  fclose (fp);

  if (!(fp = fopen (CA_KEY, "r"))) INTERR();
  if (!(CApkey = PEM_read_PrivateKey (fp, NULL, NULL, keypass))) INTERR();
  fclose (fp);


  /* ---- Create new certificate --- */

  if (!(cert = X509_new ())) INTERR();

  /* set version */
  if (!X509_set_version(cert,2)) INTERR();

  /* set issuer and subject name  */

  if (!(name = X509_get_subject_name (CAcert))) INTERR();
  if (X509_set_issuer_name (cert, name) != 1) INTERR();

  if (!(name = X509_REQ_get_subject_name (req))) ERR("REQ: No subject name");
  if (X509_set_subject_name (cert, name) != 1) INTERR();

  /* set public key */
  if (X509_set_pubkey (cert, pkey) != 1) INTERR();

  /* set duration */
  reqlt = reqlt * 60*60*24;  /* lifetime in seconds */
  if (reqlt==0) reqlt = CA_DEF_LIFETIME;
  if (reqlt<CA_MIN_LIFETIME) reqlt = CA_MIN_LIFETIME;
  if (reqlt>CA_MAX_LIFETIME) reqlt = CA_MAX_LIFETIME;
  if (!(X509_gmtime_adj (X509_get_notBefore (cert), 0))) INTERR();
  if (!(X509_gmtime_adj (X509_get_notAfter (cert), reqlt))) INTERR();

  /* Add extensions */

  /* Setup the altname. 
     Subject cn plus maybe request extension dns names */

  i = X509_NAME_get_index_by_NID(name, NID_commonName, -1);
  if (i<0) {
    fprintf (stderr, "INT: getting cn from req\n");
    INTERR();
  }
  nent = X509_NAME_get_entry(name, i);
  if (!nent) {
    fprintf (stderr, "INT: getting ent %d from req\n", i);
    INTERR();
  }
  nstr=X509_NAME_ENTRY_get_data(nent);
  if (!nstr) {
    fprintf (stderr, "INT: getting data from ent %d\n", i);
    INTERR();
  }

  /* NB.  uni2asc becomes OPENSSL_uni2asc in openssl 1.0 **/

  if (nstr->type==V_ASN1_BMPSTRING) {
     name_cn = uni2asc(nstr->data, nstr->length);
  } else {
     name_cn = OPENSSL_malloc(nstr->length+1);
     memcpy(name_cn, nstr->data, nstr->length);
     name_cn[nstr->length] = '\0';
  }
  
  alt_names = (char*) malloc(strlen(name_cn)+8);
  strcpy(alt_names,"DNS:");
  strcat(alt_names, name_cn);

  /* add any alt dns names from the request extensions */

  req_exts = X509_REQ_get_extensions(req);
  if ((flags&CA_FL_ALTNAMDNS) && req_exts) {
    for (i=0; req_exts && (i<sk_X509_EXTENSION_num(req_exts)); i++) {
     int nid;
     ext = sk_X509_EXTENSION_value(req_exts, i);
     obj = X509_EXTENSION_get_object(ext);
     nid = OBJ_obj2nid(obj);
     if (nid == NID_subject_alt_name) {
        STACK_OF(GENERAL_NAME) *alt;
        GENERAL_NAME *an;
        alt = X509V3_EXT_d2i(ext);
        if (alt) {
           int j,anl;
           char *ant;
           for (j=0;j<sk_GENERAL_NAME_num(alt);j++) {
              an = sk_GENERAL_NAME_value(alt, j);
              if (an->type != GEN_DNS) continue;
              ant = ASN1_STRING_data(an->d.ia5);
              anl = ASN1_STRING_length(an->d.ia5);
              if (!strcmp(name_cn,ant)) continue;
              alt_names = (char*) realloc(alt_names, strlen(alt_names)+anl+8);
              strcat(alt_names, ", DNS:");
              strcat(alt_names, ant);
           }
        }
     }
    }
  }
  OPENSSL_free(name_cn);


  X509V3_set_ctx (&ctx, CAcert, cert, NULL, NULL, 0);

  while (exttxt->key) {
      char *v = exttxt->value;
      if (!strcmp(v,"-cn-")) v = alt_names;
      if (!(ext=X509V3_EXT_conf(NULL,&ctx,exttxt->key, v))) {
	  fprintf (stderr, "INT: making ext \"%s = %s\"\n",
		   exttxt->key, v);
	  INTERR();
	}
      if (!X509_add_ext (cert, ext, -1))
	{
	  fprintf (stderr, "INT: adding ext \"%s = %s\"\n",
		   exttxt->key, exttxt->value);
	  INTERR();
	}
      X509_EXTENSION_free (ext);
      exttxt++;
  }


  /* get, update, and set the serial number */

  fp = fopen(CA_SERIAL,"r+");
  if (lockf(fileno(fp), LOCK_EX, 0)) INTERR();
  fscanf(fp,"%x",&serial);
  rewind(fp);
  fprintf(fp,"%06x",serial+1);
  lockf(fileno(fp), LOCK_UN, 0);
  fclose(fp);
  ASN1_INTEGER_set (X509_get_serialNumber (cert), serial);

  /* sign the certificate */
  if (EVP_PKEY_type (CApkey->type) == EVP_PKEY_DSA) digest = EVP_dss1 ();
  else if (EVP_PKEY_type (CApkey->type) == EVP_PKEY_RSA) digest = EVP_sha1 ();
  else INTERR();
  if (!(X509_sign (cert, CApkey, digest))) INTERR();

  update_db(cert);

  /* return the signed certificate */

   out = BIO_new(BIO_s_mem());
   if (out == NULL) INTERR();

   ret = BIO_get_mem_ptr(out, &outbuf);
   ret=PEM_write_bio_X509(out, cert);

   rettxt = (char*) malloc(outbuf->length+1);
   memcpy(rettxt, outbuf->data, outbuf->length);
   rettxt[outbuf->length] = '\0';

/* Common bailout for errors */

err:

   /* Free all buffers and etc */

   *crttxt = rettxt;

   if (ret) return (serial);
   return (0);
}
