/* C&C CA client.

   Contacts remote server to sign requests, create CRLs

   by Fox, February 2003

*/


#include <stdio.h>
#include <string.h>


#include "webtpl.h"
#include "caclient.h"
#include "util.h"

int  debug = 0;

/* The errors should do something html like, but they shouldn't happen. */

void handle_error(const char *file, int lineno, const char *msg)
{
    fprintf(stderr, "** %s:%i %s\n", file, lineno, msg);
    ERR_print_errors_fp(stderr);
    exit(-1);
}
#define ERR(msg) fprintf(stderr,"ERR: %s:%d %s\n", __FILE__,__LINE__,msg)

/* SSL setup and connection tools */

int verify_callback(int ok, X509_STORE_CTX *store)
{
    char data[256];
 
    if (!ok) {
      X509 *cert = X509_STORE_CTX_get_current_cert(store);
      int  depth = X509_STORE_CTX_get_error_depth(store);
      int  err = X509_STORE_CTX_get_error(store);
 
      fprintf(stderr, "-Error with certificate at depth: %i\n", depth);
      X509_NAME_oneline(X509_get_issuer_name(cert), data, 256);
      fprintf(stderr, "  issuer   = %s\n", data);
      X509_NAME_oneline(X509_get_subject_name(cert), data, 256);
      fprintf(stderr, "  subject  = %s\n", data);
      fprintf(stderr, "  err %i:%s\n", err, X509_verify_cert_error_string(err));
    }
    return ok;
}

SSL_CTX *setup_client_ctx(void)
{
    SSL_CTX *ctx;

    ctx = SSL_CTX_new(SSLv23_method(  ));
    if (SSL_CTX_load_verify_locations(ctx, path_name(CA_CERT), NULL) != 1)
        ERR("Error loading CA file and/or directory");
    if (SSL_CTX_set_default_verify_paths(ctx) != 1)
        ERR("Error loading default CA file and/or directory");
    if (SSL_CTX_use_certificate_chain_file(ctx, path_name(CL_CERT)) != 1)
        ERR("Error loading certificate from file");
    if (SSL_CTX_use_PrivateKey_file(ctx, path_name(CL_CERT), SSL_FILETYPE_PEM) != 1)
        ERR("Error loading private key from file");
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, verify_callback);
    SSL_CTX_set_verify_depth(ctx, 1);
    return ctx;
}


void init_OpenSSL(void)
{
    if (!SSL_library_init())
    {
        ERR("OpenSSL initialization failed!");
    }
    SSL_load_error_strings();
}


/* I know the server's cert uses the CN field for this */

long cert_cn_check(SSL *ssl, char *cn)
{
   X509      *cert;
   X509_NAME *subj;
   char      data[256];
   int       extcount;
   int       ok = 0;
 
   if (!(cert = SSL_get_peer_certificate(ssl)) || !cn) goto err;

 
   if (!ok && (subj = X509_get_subject_name(cert)) &&
        X509_NAME_get_text_by_NID(subj, NID_commonName, data, 256) > 0) {
      data[255] = 0;
      if (strcasecmp(data, cn) != 0) goto err;
   }
 
   X509_free(cert);
   return SSL_get_verify_result(ssl);
 
err:
    if (cert) X509_free(cert);
    return X509_V_ERR_APPLICATION_VERIFICATION;
}


/* This only works on linux */
void seed_prng(void)
{
  RAND_load_file("/dev/urandom", 1024);
}


/* ------------------------------------------------- */


int timeout = 30;
FILE *binfile = NULL;


/* password used by the CA's key */

char *kpw()
{
  static char pw[CA_KPW_LEN+1];
  FILE *f;
  char *n;
  pw[0] = '\0';
  f = fopen(path_name(CA_PW_FILE),"r");
  fgets(pw, CA_KPW_LEN, f);
  fclose(f);
  if (n=strchr(pw,'\n')) *n = '\0';
  return (pw);
}

/* These defaults can be altered before calling ca_cmd */
char *cahost = CA_HOST;
char *caport = CA_PORT;


int ca_cmd(uint32_t cmd, char *id, char *arg, uint32_t ext, uint32_t flags, char **rettxt) 
{

   int err;
   CmdHdr_ cmdhdr;
   RspHdr_ rsphdr;
 
    BIO     *conn;
    SSL     *ssl;
    SSL_CTX *ctx;

   char *caresp = NULL;
   uint32_t castatus = 0;
   struct timeval tv;
   char *s, *e, *p;
   int sslerr;
   int content_length = 0;
   uint32_t argl = 0;
   uint32_t rspl;
   char caserver[256];

   if (arg) argl = strlen(arg);

   sprintf(caserver,"%s:%s", cahost, caport);
   cmdhdr.cmd = htonl(cmd);
   cmdhdr.ext = htonl(ext);
   cmdhdr.flags = htonl(flags);
   strcpy(cmdhdr.id, id);
   strcpy(cmdhdr.kpw, kpw());
   cmdhdr.len = htonl(argl);

   /* Setup ssl connection */

    init_OpenSSL(  );
    seed_prng(  );

    ctx = setup_client_ctx(  );

    conn = BIO_new_connect(caserver);
    if (!conn) ERR("Error creating connection BIO");

    if (BIO_do_connect(conn)<=0) {
      ERR("Error connecting to server");
      ERR_print_errors_fp(stderr);
    }

    ssl = SSL_new(ctx);
    SSL_set_bio(ssl, conn, conn);
    if (SSL_connect(ssl) <= 0) ERR("Error connecting SSL object");
    if ((err = cert_cn_check(ssl, SV_CN)) != X509_V_OK)
    {
        fprintf(stderr, "-Error: peer certificate: %s\n",
                X509_verify_cert_error_string(err));
        ERR("Error checking cn after connection");
    }
    /* fprintf(stderr, "SSL Connection opened\n"); */

   /* Send the command */

   err = SSL_write(ssl, (char*)&cmdhdr, sizeof(CmdHdr_));
   if (err <= 0) return 1;
   int sz = sizeof(CmdHdr_);
    fprintf(stderr, "wrote %d header bytes ok\n", sz);

   if (argl) {
     err = SSL_write(ssl, arg, argl);
     if (err <= 0) return 3;
    fprintf(stderr, "wrote %d aux bytes ok\n", argl);
   }

   sz = sizeof(RspHdr_);
    fprintf(stderr, "attempt read %d bytes\n", sz);
   err = SSL_read(ssl, (char*)&rsphdr, sizeof(RspHdr_));

     switch (SSL_get_error(ssl, err)) {

         case SSL_ERROR_NONE:   /* record completely read */
             fprintf(stderr, "%d: error_none\n", err);
             break;

         case SSL_ERROR_ZERO_RETURN:  /* eof */
             fprintf(stderr, "%d: eof\n", err);
             break;

         case SSL_ERROR_WANT_WRITE:   /* renegotation in progress probably */
             fprintf(stderr, "%d: want write\n", err);
             break;

         case SSL_ERROR_WANT_READ:    /* record not completely received */
             fprintf(stderr, "%d: want read\n", err);
             break;

         case SSL_ERROR_SYSCALL:
             fprintf(stderr, "%d: syscall\n", err);
             break;
         case SSL_ERROR_SSL:
             fprintf(stderr, "%d: ssl err\n", err);
             break;

         default:
             fprintf(stderr, "%d: unknown\n", err);
             break;

     }

   if (err <= 0) {
      ERR("Error reading header");
      ERR_print_errors_fp(stderr);
       return (5);
   }
   castatus = ntohl(rsphdr.rsp);
   rspl = ntohl(rsphdr.len);

   if (rspl>=(cmd==CA_GETCRL?CA_MAXCRL:CA_MAXARG)) {
      return (7);
   }

   caresp = (char*) malloc(rspl+1);
   err = SSL_read(ssl, caresp, rspl);
   if (err <= 0) {
       return (9);
   }
   caresp[rspl] = '\0';

   /* printf ("len=%d, cert=%s\n", rspl, caresp); */
   
   *rettxt = caresp;
   return (castatus);
}


